python tests/artifact_store_test.py
python tests/cloud_worker_test.py            
python tests/fs_tracker_test.py        
python tests/git_util_test.py                
python tests/gpu_util_test.py                
python tests/http_provider_hosted_test.py   
python tests/http_provider_test.py          
python tests/hyperparam_test.py              
python tests/local_worker_test.py           
python tests/model_test.py                   
python tests/model_util_test.py             
python tests/providers_test.py               
python tests/queue_test.py                  
python tests/remote_worker_test.py          
python tests/runner_test.py                  
python tests/serving_test.py                 
python tests/util_test.py
           
